//**********************************************************************************************************\
//
//	pak_files.c:	Pak files
//
//**********************************************************************************************************/

#include "project.h"

// Pak file structure
// 1. Header
// 2. File bytes one after another
// 3. Directory

typedef struct
{
	char		id[4];
	int			dirofs;
	int			dirlen;
} packheader_t;

typedef struct
{
	char		name[56];
	int			filepos;
	int			filelen;
} packfile_t;

typedef struct pak_access_s
{
	packheader_t	header;
	packfile_t		directory[4096];

	char			fileName[MAX_PATH_IS_1024];
	fbool			stripPath; // Not implemented yet

	int				headerSize;
	int				totalFiles;
	int				dataBytesWritten;
	FILE*			fileHandle;
} pak_access_t;



struct pak_access_s* PakFile_Build_Start (const char* outPakFileName)
{
	pak_access_t*	myPack = Memory_calloc (sizeof(pak_access_t), 1, "In-Memory Pack File Struct");	

	// Populate it
	StringLCopy (myPack->fileName, outPakFileName);

	myPack->headerSize			= sizeof(myPack->header);
	myPack->totalFiles			= 0;
	myPack->dataBytesWritten	= 0;

	myPack->fileHandle = fopen (myPack->fileName, "wb");

	if (!myPack->fileHandle)
	{
		// Deallocate and return NULL
		myPack = Memory_free(myPack);
		return NULL;
	}

	// Write the header with invalid data.  Don't worry we will be updating this later.
	File_Safe_Write (myPack->fileHandle, &myPack->header, myPack->headerSize);
		
	return myPack;

}

// Name options ... TopDir or Bare 
fbool PakFile_Build_Add_File (struct pak_access_s* myPack, const char* newFileName)
{
	int i = myPack->totalFiles;

	if (myPack->totalFiles >= 4096)
	{
		printf ("Pak directory full with %i files", myPack->totalFiles);
		return False;
	}

	if (strlen(String_SkipPath(newFileName)) > MAX_PAK_FILENAME_IS_55)
	{
		printf ("File name is too long '%s' with %i chars (MAX %i)\n", newFileName, strlen(String_SkipPath(newFileName)), MAX_PAK_FILENAME_IS_55);
		return False;
	}

	StringLCopy (myPack->directory[i].name, String_SkipPath(newFileName));
	
	myPack->directory[i].filepos = File_GetCursor		  (myPack->fileHandle);  // LittleLong
	myPack->directory[i].filelen = File_Append_BinaryFile (myPack->fileHandle, newFileName); // This writes the file.   LittleLong
	
	myPack->dataBytesWritten += myPack->directory[i].filelen;
	myPack->totalFiles ++;

	return True;
}

struct pak_access_s* PakFile_Build_Finish (struct pak_access_s* myPack)
{
	// Store the offset then write the directory.  Data at pakDirectory with entry size of entry 0 times filecount
	int directoryStart			= File_GetCursor (myPack->fileHandle);	
	int directoryEntrySize		= sizeof(myPack->directory[0]);
	int directoryByteLength		= directoryEntrySize * myPack->totalFiles;

	File_Safe_Write (myPack->fileHandle, &myPack->directory, directoryByteLength);
	
	// Construct final header
	myPack->header.id[0] = 'P';
	myPack->header.id[1] = 'A';
	myPack->header.id[2] = 'C';
	myPack->header.id[3] = 'K';
	myPack->header.dirofs = LittleLong (directoryStart);
	myPack->header.dirlen = LittleLong (directoryByteLength);

	// Go back to the beginning and rewrite the header

	File_SetCursor_ToStart (myPack->fileHandle);
	File_Safe_Write (myPack->fileHandle, &myPack->header, myPack->headerSize);

	// close the file
	fclose (myPack->fileHandle);

	// Free the memory (this should effective return NULL, since Memory_free returns NULL)
	return Memory_free (myPack);
}


typedef struct
{ 
	byte		magic[4];           // "WAD2", Name of the new WAD format
	int			numentries;         // Number of entries
	int			diroffset;          // Position of WAD directory in file
} wadheader_t;

typedef struct
{ 
	int			offset;				// Position of the entry in WAD
	int			wadsize;            // Size of the entry in WAD file
	int			memsize;            // Size of the entry in memory
	char		type;               // type of entry, should be 'D' (0x44) MipTex
	char		unused[3];          // Never compressed, so why bother
	char		name[16];           // 1 to 16 characters, '\0'-padded
} wadentry_t;

typedef struct wad_access_s
{
	wadheader_t		header;
	wadentry_t		directory[4096];

	char			fileName[MAX_PATH_IS_1024];
	fbool			stripPath; // Not implemented yet
	char			entryChar;

	int				headerSize;
	int				totalFiles;
	int				dataBytesWritten;
	FILE*			fileHandle;
} wad_access_t;



static struct wad_access_s* WADFile_Build_Start (const char* outWADFileName)
{
	wad_access_t*	myWAD = Memory_calloc (sizeof(pak_access_t), 1, "In-Memory Pack File Struct");	

	// Populate it
	StringLCopy (myWAD->fileName, outWADFileName);

	myWAD->headerSize			= sizeof(myWAD->header);
	myWAD->totalFiles			= 0;
	myWAD->dataBytesWritten	= 0;

	myWAD->fileHandle = fopen (myWAD->fileName, "wb");

	if (!myWAD->fileHandle)
	{
		// Deallocate and return NULL
		myWAD = Memory_free(myWAD);
		return NULL;
	}

	// Write the header with invalid data.  Don't worry we will be updating this later.
	File_Safe_Write (myWAD->fileHandle, &myWAD->header, myWAD->headerSize);
		
	return myWAD;

}

#define isWAD3(myWad) (myWad->entryChar == 'C')
struct wad_access_s* WAD2File_Build_Start (const char* outWADFileName)
{
	wad_access_t* myWAD = WADFile_Build_Start (outWADFileName);

	if (myWAD)
		myWAD->entryChar = 'D'; // WAD2 uses D

	return myWAD;
}

struct wad_access_s* WAD3File_Build_Start (const char* outWADFileName)
{
	wad_access_t* myWAD = WADFile_Build_Start (outWADFileName);

	if (myWAD)
		myWAD->entryChar = 'C'; // WAD3 uses C

	return myWAD;
}

extern byte quakePalette [768];
// Name options ... TopDir or Bare 
fbool WADFile_Build_Add_File (struct wad_access_s* myWAD, const char* newFileName)
{
	char	newTextureName[MAX_PATH_IS_1024];
	byte	pad[2] = {0, 1};
	int		i = myWAD->totalFiles;
	int		strlenTextureName;

	if (myWAD->totalFiles >= 4096)
	{
		printf ("WAD2 directory full with %i files", myWAD->totalFiles);
		return False;
	}

	// Prepare texture name.  #water becomes *water
	StringLCopy (newTextureName, String_SkipPath(newFileName));
	String_Edit_StripExtension (newTextureName);
	
	if (newTextureName[0] = '#') newTextureName[0] = '*'; // Fix warp texture names

	strlenTextureName = strlen(newTextureName);

	if (strlenTextureName > MAX_WAD_TEXTURE_NAME_LENGTH_IS_15)
	{
		printf ("File name is too long '%s' with %i chars (MAX %i)\n", newTextureName, strlenTextureName, MAX_WAD_TEXTURE_NAME_LENGTH_IS_15);
		return False;
	}
	
	myWAD->directory[i].offset = File_GetCursor		  (myWAD->fileHandle);  // LittleLong
	
	{
		int startingBytesWritten = myWAD->dataBytesWritten;
		int netsize;

		myWAD->dataBytesWritten    += File_Append_BinaryFile (myWAD->fileHandle, newFileName); // This writes the file.   LittleLong

		if (isWAD3(myWAD))
		{
			myWAD->dataBytesWritten    +=  File_Safe_Write (myWAD->fileHandle, pad,		     sizeof(pad));			// 2
			myWAD->dataBytesWritten    +=  File_Safe_Write (myWAD->fileHandle, quakePalette, sizeof(quakePalette));	// 768
			
			// 4 byte align if WAD3, bumping up wadsize, memsize and dataBytesWritten
			while (myWAD->dataBytesWritten & 3)
				myWAD->dataBytesWritten += File_Safe_Write (myWAD->fileHandle, pad,				1);

		}

		// wadsize is disk size, which is always memsize ... right?
		myWAD->directory[i].wadsize = myWAD->directory[i].memsize = netsize = myWAD->dataBytesWritten - startingBytesWritten;
	}

//	myWAD->directory[i].type	 = 'D'; // 0x44 MipTex   WAD2
//	myWAD->directory[i].type	 = 'C'; // 0x43 MipTex   WAD3

	myWAD->directory[i].type	 = myWAD->entryChar;

	StringLCopy (myWAD->directory[i].name, newTextureName);

	myWAD->totalFiles ++;

	return True;
}

struct wad_access_s* WADFile_Build_Finish (struct wad_access_s* myWAD)
{
	// Store the offset then write the directory.  Data at pakDirectory with entry size of entry 0 times filecount
	int directoryStart			= File_GetCursor (myWAD->fileHandle);	
	int directoryEntrySize		= sizeof(myWAD->directory[0]);
	int directoryByteLength		= directoryEntrySize * myWAD->totalFiles;

	File_Safe_Write (myWAD->fileHandle, &myWAD->directory, directoryByteLength);
	
	// Construct final header
	myWAD->header.magic[0] = 'W';
	myWAD->header.magic[1] = 'A';
	myWAD->header.magic[2] = 'D';
	myWAD->header.magic[3] = isWAD3(myWAD) ?  '3' : '2';
	myWAD->header.numentries = LittleLong (myWAD->totalFiles);
	myWAD->header.diroffset = LittleLong (directoryStart);

	// Go back to the beginning and rewrite the header

	File_SetCursor_ToStart (myWAD->fileHandle);
	File_Safe_Write (myWAD->fileHandle, &myWAD->header, myWAD->headerSize);

	// close the file
	fclose (myWAD->fileHandle);

	// Free the memory (this should effective return NULL, since Memory_free returns NULL)
	return Memory_free (myWAD);
}
